#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 30.04.20
@author: felix
"""
from datetime import datetime
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Type
from typing import Union, Tuple

import pytest

from strongtyping.strong_typing import match_typing, TypeMisMatch


def test_func_without_typing():
    @match_typing
    def func_a(a, b):
        return True

    assert func_a(1, '2') is True


def test_func_keeps_docstring():
    @match_typing
    def func_a(a, b):
        """
        Test abracadabra
        """
        return True

    assert 'Test abracadabra' == func_a.__doc__.strip()
    assert 'func_a' == func_a.__name__


def test_func_with_one_typing_param():
    @match_typing
    def func_a(a, b: int):
        return True

    assert func_a(1, 2) is True


def test_func_with_multiple_typing_param():
    @match_typing
    def func_a(a: str, b: int, c: list):
        return True

    assert func_a('1', 2, [i for i in range(5)])


def test_func_with_union_typing_param():
    @match_typing
    def func_a(a: Union[str, int], b: int, c: Union[list, tuple]):
        return True

    assert func_a(3, 4, ('capacious', 'extremis'))


def test_func_with_tuple_typing():
    @match_typing
    def func_a(a: Tuple[str, str, str]):
        return True

    assert func_a(('Harmonia', 'Nectere', 'Passus'))


def test_func_raise_error_incorrect_parameters_less():
    @match_typing
    def func_a(a: Tuple[str, str, str]):
        return True

    with pytest.raises(TypeMisMatch):
        func_a(a=('Oculus', 'Reparo'))


def test_func_raise_error_incorrect_parameters_much():
    @match_typing
    def func_a(a: Tuple[str, str, str]):
        return True

    with pytest.raises(TypeMisMatch):
        func_a(('Peskipiksi', 'Pesternomi', 'Petrificus', 'Totalus'))


def test_func_with_own_union_type():
    class A:
        pass

    class B:
        pass

    MyType = Tuple[Union[str, int], Union[list, tuple], Union[A, B]]

    @match_typing
    def func_a(a: MyType):
        return True

    assert func_a(a=('Lumos', [1, 2, 3], A()))


def test_func_with_own_union_type():
    class A:
        pass

    class B:
        pass

    MyType = Tuple[Union[str, int], Union[list, tuple], Union[A, B]]

    @match_typing
    def func_a(a: MyType):
        return True

    assert func_a(a=('Lumos', [1, 2, 3], A()))


def test_decorated_class_method():

    class A:
        @match_typing
        def func_a(self, a: int):
            return True

        @staticmethod
        @match_typing
        def func_b(b: str):
            return True

        @classmethod
        @match_typing
        def func_c(cls, c: str):
            return True

    a = A()
    assert a.func_a(2) is True
    assert a.func_b('test success')
    assert A.func_c('test success')


def test_use_different_exception():

    @match_typing(excep_raise=TypeError)
    def func_a(a: str):
        return True

    @match_typing(excep_raise=ValueError)
    def func_b(a: str):
        return True

    with pytest.raises(TypeError):
        func_a(12)

    with pytest.raises(ValueError):
        func_b(22)


def test_use_own_exception():

    class MyException(Exception):
        pass

    @match_typing(excep_raise=MyException)
    def func_a(a: str):
        return True

    with pytest.raises(MyException):
        func_a(12)


def test_use_str_repr_as_type():

    class A:
        @match_typing
        def func_a(self, a: 'A'):
            return True

    class Foo:
        pass

    with pytest.raises(TypeMisMatch):
        b = A()
        b.func_a(Foo())

    b = A()
    assert b.func_a(A())


def test_second_pos_arg_hinted():

    @match_typing
    def func_b(a, b: int):
        return f"{a}, {b}"

    assert func_b(1, 2) == "1, 2"

    assert func_b("1", 2) == "1, 2"  # shouldn't raise a TypeMisMatch

    with pytest.raises(TypeMisMatch):
        assert func_b(1, "2") == "1, 2"  # Should raise a TypeMisMatch

    with pytest.raises(TypeMisMatch):
        assert func_b("1", "2") == "1, 2"


def test_with_lists():

    @match_typing
    def func_b(a, b: List):
        return f"{len(a)}, {len(b)}"

    # assert func_b([1, 2], ['a', 'b', 'c']) == '2, 3'

    with pytest.raises(TypeMisMatch):
        func_b([1, 2], ('a', 'b', 'c')) == '2, 3'

    @match_typing
    def func_c(a: list, b: List[str]):
        return f"{len(a)}, {len(b)}"

    assert func_c([1, 2], ['a', 'b', 'c']) == '2, 3'

    with pytest.raises(TypeMisMatch):
        func_c([1, 2], ('a', 'b', 'c')) == '2, 3'

    with pytest.raises(TypeMisMatch):
        func_c((1, 2), ['a', 'b', 'c']) == '2, 3'

    with pytest.raises(TypeMisMatch):
        func_c([1, 2], [1, 2, 3]) == '2, 3'

    @match_typing
    def func_c(a: list, b: List[int]):
        return f"{len(a)}, {len(b)}"

    assert func_c([1, 2], [1, 2, 3]) == '2, 3'

    @match_typing
    def func_d(a: list, b: List[Union[int, str]]):
        return f"{len(a)}, {len(b)}"

    assert func_d([1, 2], [1, 2, '3', '4']) == '2, 4'


def test_lists_with_unions():

    @match_typing
    def func_e(a: List[Union[str, int]], b: List[Union[str, int, tuple]]):
        return f'{len(a)}-{len(b)}'

    assert func_e([1, '2', 3, '4'], [5, ('a', 'b'), '10'])

    with pytest.raises(TypeMisMatch):
        func_e([5, ('a', 'b'), '10'], [1, '2', 3, datetime.date])


def test_with_any():
    @match_typing
    def func_a(a: Any, b: any):
        return f'{a}-{b}'

    assert func_a(2, '2') == '2-2'
    assert func_a([], ()) == '[]-()'
    assert func_a(datetime, set())


def test_with_optional():

    @match_typing
    def func_a(a: Optional[str], b: Optional[int]):
        return f'{a}-{b}'

    assert func_a(None, None) == 'None-None'
    assert func_a('2', 1) == '2-1'

    with pytest.raises(TypeMisMatch):
        func_a(1, '2')


def test_with_dict():

    @match_typing
    def func_a(a: Dict[str, int], b: Dict[Tuple[str, str], int]):
        return f'{a}-{b}'

    assert func_a({'a': 5, 'b': 2},
                  {('hello', 'world'): 10,
                   ('foo', 'bar'): 6}) == "{'a': 5, 'b': 2}-{('hello', 'world'): 10, ('foo', 'bar'): 6}"

    with pytest.raises(TypeMisMatch):
        func_a({'a': 5, 'b': 2}, {'helloworld': 10, 'foobar': 6})

    with pytest.raises(TypeMisMatch):
        func_a({'a': 5, 'b': 2}, {('hello', 'world'): '2', ('foo', 'bar'): '9'})

    with pytest.raises(TypeMisMatch):
        func_a({'a': 5, 'b': '2'}, {('hello', 'world'): 12, ('foo', 'bar'): 19})

    with pytest.raises(TypeMisMatch):
        func_a({'a': 5, 'b': 2}, {('hello', 'world'): 12, ('foo', 'bar'): '19'})


def test_with_dict_2():
    @match_typing
    def func_b(a: Dict[str, int], b: Dict[Tuple[Tuple[Tuple[str, str], str], str], int]):
        return True

    @match_typing
    def func_c(a: Dict[Union[str, int], int]):
        return True

    assert func_b({'a': 1}, {((('fbar', 'fbar'), 'foo'), 'bar'): 2020})
    assert func_c({'b': 2, 34: 313})

    with pytest.raises(TypeMisMatch):
        assert func_b({'a': 1}, {((('fbar', 1), 'foo'), 'bar'): 2020})

    with pytest.raises(TypeMisMatch):
        assert func_c({'b': 2, 34: 'foo'})


def test_with_set():
    @match_typing
    def func_a(a: Set[Union[str, int]], b: Set[int]):
        return True

    assert func_a({'A', 2, 'b', 4, 'c'}, {1, 2, 3, 4})

    with pytest.raises(TypeMisMatch):
        func_a({'A', 2, 'b', 4, 'c'}, {'A', 2, 'b', 4, 'c'})


def test_with_type():

    class NoUser:
        def __repr__(self):
            return 'NoUser'

    class User:
        def __repr__(self):
            return 'User'

    class BasicUser(User):
        pass

    class ProUser(User):
        pass

    class TeamUser(BasicUser, ProUser):
        pass

    @match_typing
    def func_a(a: Type[User]):
        _a = a()
        return repr(_a)

    @match_typing
    def func_b(a: Type[User], b: Type[Union[BasicUser, ProUser]]):
        _a, _b = a(), b()
        return repr(_a), repr(_b)

    assert func_a(User) == 'User'
    assert func_b(User, TeamUser) == ('User', 'User')

    with pytest.raises(TypeMisMatch):
        func_a(NoUser)

    with pytest.raises(TypeMisMatch):
        func_a(NoUser, User)


def test_with_iterator():

    class Fibonacci:
        def __init__(self, limit):
            self.limit = limit
            self.count = 0
            self.values = []

        def __iter__(self):
            return self

        def __next__(self):
            self.count += 1
            if self.count > self.limit:
                raise StopIteration

            if len(self.values) < 2:
                self.values.append(1)
            else:
                self.values = [self.values[-1], self.values[-1] + self.values[-2]]
            return self.values[-1]

    class NonFibonacci:
        def __init__(self, limit):
            self.limit = limit
            self.count = 0
            self.values = []

    @match_typing
    def func_a(a: Iterator):
        return [f for f in a]

    assert func_a(Fibonacci(5)) == [1, 1, 2, 3, 5]

    with pytest.raises(TypeMisMatch):
        func_a(NonFibonacci(5))


def test_with_callable():

    def dummy_func(a: int, b: str, c: Union[str, int]) -> str:
        return 'success'

    def fail_func(a: int, b: str, c: Union[str, int]) -> int:
        return 42

    @match_typing
    def func_a(a: Callable[[int, str, Union[str, int]], str]):
        return True

    assert func_a(dummy_func)
    with pytest.raises(TypeMisMatch):
        func_a(fail_func)


if __name__ == '__main__':
    pytest.main(['-vv', '-s'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 28.04.20
@author: felix
"""
import functools
import inspect
from itertools import zip_longest
import typing

from strongtyping.cached_set import CachedSet


class TypeMisMatch(AttributeError):
    def __init__(self, message):
        super().__init__()
        print(message)


typing_base_class = typing._GenericAlias if hasattr(typing, '_GenericAlias') else typing.GenericMeta


def get_possible_types(typ_to_check) -> typing.Union[tuple, None]:
    typ_check_args = typ_to_check.__args__
    if typ_check_args is not None:
        return tuple(typ for typ in typ_check_args if not isinstance(typ, typing.TypeVar))


def get_fillvalue(typ_to_check: any, return_val: any) -> typing.Union[str, int, None]:
    if hasattr(typ_to_check, '_name'):
        return return_val if typ_to_check._name == 'List' else None


def get_origins(typ_to_check: any) -> tuple:
    origin = None
    if hasattr(typ_to_check, '__origin__') or hasattr(typ_to_check, '__orig_bases__'):
        origin = typ_to_check.__origin__ if typ_to_check.__origin__ is not None else typ_to_check.__orig_bases__
    return origin, origin._name if hasattr(origin, '_name') else\
        typ_to_check._name if hasattr(typ_to_check, '_name') else f'{typ_to_check}'


def check_typing_dict(arg: typing.Any, possible_types: tuple):
    if not isinstance(arg, dict):
        return False
    key, val = possible_types
    if hasattr(key, '__origin__'):
        result_key = all(check_type(a, key) for a in arg.keys())
    else:
        result_key = all(isinstance(k, key) for k in arg.keys())
    if hasattr(val, '__origin__'):
        result_val = all(check_type(a, val) for a in arg.values())
    else:
        result_val = all(isinstance(v, val) for v in arg.values())
    return result_key and result_val


def checking_typing_set(arg: typing.Any, possible_types: tuple):
    pssble_type = possible_types[0]
    return all(check_type(argument, pssble_type) for argument in arg)


def checking_typing_type(arg: typing.Any, possible_types: tuple):
    if not hasattr(arg, '__mro__'):
        return False
    arguments = arg.__mro__
    return any(check_type(arguments, possible_type, mro=True) for possible_type in possible_types)


def checking_typing_union(arg: typing.Any, possible_types: tuple, mro):
    if mro:
        return any(pssble_type in arg for pssble_type in possible_types)
    try:
        return isinstance(arg, possible_types)
    except TypeError:
        return all(check_type(arg, typ) for typ in possible_types)


def checking_typing_iterator(arg: typing.Any):
    return hasattr(arg, '__iter__') and hasattr(arg, '__next__')


def checking_typing_callable(arg: typing.Any, possible_types: tuple):
    insp = inspect.signature(arg)
    return_val = insp.return_annotation == possible_types[-1]
    params = insp.parameters
    return return_val and all(p.annotation == pt for p, pt in zip(params.values(), possible_types))


def check_type(argument, type_of, mro=False):
    check_result = True
    if type_of is not None:
        origin, origin_name = get_origins(type_of)
        if 'any' in origin_name.lower():
            return check_result
        if isinstance(type_of, typing_base_class):
            if hasattr(type_of, '__origin__'):
                possible_types = get_possible_types(type_of)

                if 'dict' in origin_name.lower():
                    return check_typing_dict(argument, possible_types)
                if 'set' in origin_name.lower():
                    return checking_typing_set(argument, possible_types)
                if 'type' in origin_name.lower():
                    return checking_typing_type(argument, possible_types)
                if 'iterator' in origin_name.lower():
                    return checking_typing_iterator(argument)
                if 'callable' in origin_name.lower():
                    return checking_typing_callable(argument, possible_types)

                if possible_types and origin_name != 'Union':
                    fillvalue = get_fillvalue(type_of, possible_types[0])
                    check_result = isinstance(argument, origin) and all(check_type(arg, typ) for arg, typ in
                                                                        zip_longest(argument, possible_types,
                                                                                    fillvalue=fillvalue)) and (
                                           len(argument) == len(type_of.__args__) or (isinstance(argument, list)))
                elif origin_name == 'Union':
                    return checking_typing_union(argument, possible_types, mro)
                else:
                    possible_type = origin[0] if isinstance(origin, (list, tuple)) else origin
                    check_result = isinstance(argument, possible_type)
            else:
                check_result = isinstance(argument, type_of.__args__)
        elif isinstance(type_of, str):
            check_result = argument.__class__.__name__ == type_of
        elif mro:
            check_result = type_of in argument
        else:
            try:
                check_result = isinstance(argument, type_of)
            except TypeError:
                check_result = isinstance(argument, type_of._subs_tree()[1:])
    return check_result


def match_typing(_func=None, *, excep_raise: Exception = TypeMisMatch, cache_size=0):
    cached_set = CachedSet(cache_size) if cache_size > 0 else None

    def wrapper(func):
        arg_names = [name for name in inspect.signature(func).parameters]
        annotations = func.__annotations__

        @functools.wraps(func)
        def inner(*args, **kwargs):

            if cached_set is not None:
                # check if func with args and kwargs was checked once before with positive result
                cached_key = (func, repr(args), repr(kwargs))
                if cached_key in cached_set:
                    return func(*args, **kwargs)

            # Thanks to Ruud van der Ham who find a better and more stable solution for check_args
            failed_arg_names = tuple(
                arg_name for arg, arg_name in zip(args, arg_names) if not check_type(arg, annotations.get(arg_name))
            )
            failed_kwarg_names = tuple(
                kwarg_name for kwarg_name, kwarg in kwargs.items() if not check_type(kwarg, annotations.get(kwarg_name))
            )

            if failed_arg_names or failed_kwarg_names:
                params = [f"{name}: {annotations[name]}" for name in failed_arg_names + failed_kwarg_names]
                msg = f'Incorrect parameters: {", ".join(params)}'
                raise excep_raise(msg)

            if cached_set is not None:
                cached_set.add(cached_key)

            return func(*args, **kwargs)

        return inner

    if _func is not None:
        return wrapper(_func)
    else:
        return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 07.06.20
@author: felix
"""
import sys
from typing import Union


class CachedSet(set):
    """
    Warning only use for caching when Memory limit is reached all items will be cleared
    """

    def __init__(self, memory_limit: Union[int, float] = 1, *args, **kwargs):
        """
        :param memory_limit: in MB
        """
        self.memory_limit = memory_limit * 1000000
        super(CachedSet, self).__init__(*args, **kwargs)

    def add(self, element: any) -> None:
        if sys.getsizeof(self) > self.memory_limit:
            self.clear()
        super().add(element)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 30.04.20
@author: felix
"""
import pathlib

from setuptools import find_packages
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()
setup(
    name="strongtyping",
    version="1.1.16",
    description="Decorator which checks whether the function is called with the correct type of parameters",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/FelixTheC/strongtyping",
    author="FelixTheC",
    author_email="fberndt87@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(exclude=['test_*']),
    python_requires='>=3.6',
    include_package_data=True,
)
